# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# wordloom

'''
Routines to help with processing text in the word loom convention:
https://github.com/OoriData/WordLoom

Word Loom is a TOML-based format for managing language text and templates
for AI/LLM applications, with support for multiple languages and metadata.
'''

import io
import tomli
import warnings


class language_item(str):
    '''
    Text or template for use with LLM tools
    Largely keeps metadata around language, template markers, etc.

    >>> from wordloom import language_item as LI
    >>> t = LI('spam', deflang='en')
    >>> t
    'spam'
    >>> t.lang
    'en'
    >>> t = LI('spam', deflang='en', altlang={'fr': 'jambon'})
    >>> t.altlang['fr']
    'jambon'
    '''

    def __new__(cls, value, deflang, altlang=None, meta=None, markers=None):
        '''
        Construct a new text item

        value - text value in the default language
        deflang - default language - made mandatory to avoid sloppy language assumptions
        altlang - dictionary of text values in alternative languages
        meta - dictionary of metadata
        markers - used to specify values that can be set, with the text value is treated as a template
        '''
        assert isinstance(value, str)
        self = super(language_item, cls).__new__(cls, value)
        self.lang = deflang  # Default language
        self.meta = meta or {}
        self.markers = markers or {}
        self.altlang = altlang or {}
        return self

    def __repr__(self):
        return u'T(' + repr(str(self)) + ')'

    def in_lang(self, lang):
        return self.altlang.get(lang)

    def clone(self, value=None, deflang=None, altlang=None, meta=None, markers=None):
        '''
        Clone the text item, with optional overrides

        >>> from wordloom import language_item as LI
        >>> t = LI('spam', deflang='en', meta={'tag': 'food'})
        >>> t, t.meta
        'spam', {'tag': 'food'}
        >>> t_cloned = t.clone(meta={'tag': 'protein'})
        >>> t_cloned, t_cloned.meta
        'spam', {'tag': 'protein'}
        >>> t_cloned = t.clone('eggs')
        >>> t_cloned, t_cloned.meta
        'spam', {'tag': 'food'}
        '''
        value = str(self) if value is None else value
        deflang = self.lang if deflang is None else deflang
        altlang = self.altlang if altlang is None else altlang
        meta = self.meta if meta is None else meta
        markers = self.markers if markers is None else markers
        return language_item(value, deflang, altlang=altlang, meta=meta, markers=markers)


# Following 2 lines are deprecated
T = language_item
text_item = language_item

LI = language_item  # Alias for language_item


# XXX Defaulting to en leaves a bit too imperialist a flavor, really
def load(fp_or_str, lang='en', preserve_key=False):
    '''
    Read a word loom and return the tables as top-level result mapping
    Loads the TOML

    Return a dict of the language items, indexed by the TOML key as well as its default language text

    fp_or_str - file-like object or string containing TOML
    lang - select only texts in this language (default: 'en')
    preserve_key - if True, the key in the TOML is preserved in each item's metadata

    Example:
    >>> import wordloom
    >>> with open('prompts.toml', mode='rb') as fp:
    ...     loom = wordloom.load(fp)
    >>> loom['system_instruction'].meta
    {'category': 'system'}
    >>> actual_text = loom['translation_prompt']
    >>> str(actual_text)
    'Translate the following text to {target_lang}: {text}'
    >>> actual_text.markers
    ['target_lang', 'text']
    '''
    # Ensure we have a file-like object
    if isinstance(fp_or_str, str):
        fp_or_str = io.BytesIO(fp_or_str.encode('utf-8'))
    elif isinstance(fp_or_str, bytes):
        fp_or_str = io.BytesIO(fp_or_str)
    # Load TOML
    loom_raw = tomli.load(fp_or_str)
    # Select text by language
    # FIXME: Only top level, for now. Presumably we'll want to support scoping
    texts = {}
    default_lang = loom_raw.get('lang', None)
    for k, v in loom_raw.items():
        if not isinstance(v, dict):
            # Skip top-level items
            continue
        if 'text' in v:
            warnings.warn('Deprecated attribute "text". Use "_" instead')
            text = v['text']
        else:
            text = v.get('_')
        if text is None: # Skip items without text
            continue
        markers = v.get('_m')
        if 'markers' in v:
            warnings.warn('Deprecated attribute "marker". Use "_m" instead')
            markers = v['markers']
        else:
            markers = v.get('_m')
        if v.get('lang') == lang or ('lang' not in v and lang == default_lang):
            altlang = {kk.lstrip('_'): vv for kk, vv in v.items() if (kk.startswith('_') and kk not in ('_', '_m'))}
            meta = {kk: vv for kk, vv in v.items() if (not kk.startswith('_') and kk not in ('text', 'markers'))}
            if preserve_key:
                meta['_key'] = k
            if k in texts:
                warnings.warn(f'Key {k} duplicates an existing item, which will be overwritten')
            texts[k] = T(text, lang, altlang=altlang, meta=meta, markers=markers)
            # Also index by literal text
            if text in texts:
                warnings.warn(
                    f'Item default language text {text[:20]} duplicates an existing item, which will be overwritten')
            texts[text] = T(text, lang, altlang=altlang, meta=meta, markers=markers)
    return texts



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
#
# SPDX-License-Identifier: Apache-2.0
# wordloom

# ruff: noqa: F401,F403

from .__about__ import __version__
from .wordloom import load, language_item, LI



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_basics.py
'''
pytest test

or

pytest test/test_basics.py
'''
# ruff: noqa: E501

import os
import sys

import pytest

import wordloom


@pytest.fixture
def SAMPLE_TOML_STR():
    # Load language material (word loom) format
    fpath = os.path.dirname(sys.modules['wordloom'].__file__)

    with open(os.path.join(fpath, 'resources/wordloom/sample.toml'), 'rb') as fp:
        toml_content = fp.read()
    return toml_content


@pytest.fixture
def SAMPLE_TOML_FP():
    # Load language material (word loom) format
    fpath = os.path.dirname(sys.modules['wordloom'].__file__)

    return open(os.path.join(fpath, 'resources/wordloom/sample.toml'), 'rb')

def test_load_fp_vs_str(SAMPLE_TOML_STR, SAMPLE_TOML_FP):
    loom1 = wordloom.load(SAMPLE_TOML_STR)
    loom2 = wordloom.load(SAMPLE_TOML_FP)
    SAMPLE_TOML_FP.close()
    assert loom1 == loom2


def test_sample_texts_check(SAMPLE_TOML_STR):
    # print(SAMPLE_TOML)
    loom = wordloom.load(SAMPLE_TOML_STR)
    # default language text is also a key
    # We now have more items: i18n examples + general-purpose examples
    # Each item appears twice: once by key, once by text value
    assert len(loom.keys()) >= 10  # At least the core items

    # Check for i18n examples
    for k in ['davinci3_instruct_system', 'hello_translated', 'i18n_context', 'write_i18n_advocacy']:
        assert k in loom.keys()

    # Check for general-purpose examples
    for k in ['system_instruction', 'code_review_prompt', 'greeting_multilang']:
        assert k in loom.keys()

    assert 'Hello' in loom.keys()

    # Check markers for specific items
    assert loom['write_i18n_advocacy'].markers == ['davinci3_instruct_system', 'i18n_context']
    assert loom['code_review_prompt'].markers == ['code_snippet']
    assert loom['davinci3_instruct_system'].lang == 'en'
    assert loom['system_instruction'].lang == 'en'

    # Check multilang item has translations
    greeting = loom['greeting_multilang']
    assert greeting.in_lang('fr') is not None
    assert greeting.in_lang('ja') is not None
    assert 'Bonjour' in greeting.in_lang('fr')

    # Default language is English
    loom1 = wordloom.load(SAMPLE_TOML_STR, lang='en')
    assert loom1 == loom

    # Test French language selection
    loom_fr = wordloom.load(SAMPLE_TOML_STR, lang='fr')
    assert 'goodbye_translated' in loom_fr.keys()
    assert 'translate_request' in loom_fr.keys()
    assert 'hardcoded_food' in loom_fr.keys()
    assert loom_fr['hardcoded_food'].lang == 'fr'
    assert loom_fr['translate_request'].markers == ['hardcoded_food']


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_i18n_integration.py
'''
Test WordLoom i18n features with OpenAI API integration

pytest test/test_i18n_integration.py
'''

import os
import sys

import pytest

import wordloom


@pytest.fixture
def i18n_loom():
    '''Load the sample loom which includes i18n examples'''
    fpath = os.path.dirname(sys.modules['wordloom'].__file__)
    with open(os.path.join(fpath, 'resources/wordloom/sample.toml'), 'rb') as fp:
        return wordloom.load(fp)


def test_multilingual_greeting(i18n_loom):
    '''Test accessing greetings in multiple languages'''
    greeting = i18n_loom['greeting_multilang']

    # Check all languages are present
    assert str(greeting) == 'Hello, how can I help you today?'
    assert 'Bonjour' in greeting.in_lang('fr')
    assert 'Hola' in greeting.in_lang('es')
    assert 'Hallo' in greeting.in_lang('de')
    assert 'こんにちは' in greeting.in_lang('ja')


def test_language_override():
    '''Test that language overrides work correctly'''
    # goodbye_translated has lang='fr' set, so we need to load with lang='fr'
    fpath = os.path.dirname(sys.modules['wordloom'].__file__)
    with open(os.path.join(fpath, 'resources/wordloom/sample.toml'), 'rb') as fp:
        loom_fr = wordloom.load(fp, lang='fr')

    goodbye = loom_fr['goodbye_translated']
    assert goodbye.lang == 'fr'
    assert str(goodbye) == 'Adieu'
    assert goodbye.in_lang('en') == 'Goodbye'


def test_template_composition_i18n(i18n_loom):
    '''Test that template composition works with i18n content'''
    template = i18n_loom['write_i18n_advocacy']

    # Check markers are preserved
    assert 'davinci3_instruct_system' in template.markers
    assert 'i18n_context' in template.markers

    # Check the template references other items
    assert '{davinci3_instruct_system}' in str(template)
    assert '{i18n_context}' in str(template)


def test_metadata_with_i18n(i18n_loom):
    '''Test that metadata is preserved on i18n items'''
    context = i18n_loom['i18n_context']

    # Check metadata
    assert 'source' in context.meta
    assert 'lionbridge.com' in context.meta['source']


def test_french_language_selection():
    '''Test loading only French language items'''
    fpath = os.path.dirname(sys.modules['wordloom'].__file__)
    with open(os.path.join(fpath, 'resources/wordloom/sample.toml'), 'rb') as fp:
        loom_fr = wordloom.load(fp, lang='fr')

    # French items should be present
    assert 'hardcoded_food' in loom_fr
    assert 'translate_request' in loom_fr
    assert 'goodbye_translated' in loom_fr

    # Check the text is in French
    assert str(loom_fr['hardcoded_food']) == 'pomme de terre'
    assert 'Comment dit-on' in str(loom_fr['translate_request'])


def test_openai_multilingual_workflow(i18n_loom, mocker):
    '''
    Test a realistic multilingual workflow with OpenAI API mock
    Simulates serving users in different languages
    '''
    # Mock the OpenAI client
    mock_client = mocker.MagicMock()
    mock_response = mocker.MagicMock()
    mock_response.choices = [mocker.MagicMock()]
    mock_response.choices[0].message.content = 'Voici votre réponse'

    mock_client.chat.completions.create.return_value = mock_response

    # Simulate user preference for French
    user_lang = 'fr'
    greeting = i18n_loom['greeting_multilang']

    # Select appropriate language
    if user_lang != 'en':
        greeting_text = greeting.in_lang(user_lang) or str(greeting)
    else:
        greeting_text = str(greeting)

    # Verify French greeting was selected
    assert 'Bonjour' in greeting_text

    # Use with mock OpenAI
    response = mock_client.chat.completions.create(  # noqa: F841
        model='gpt-4',
        messages=[
            {'role': 'system', 'content': greeting_text},
            {'role': 'user', 'content': 'J\'ai besoin d\'aide'}
        ]
    )

    # Verify the call was made with French greeting
    assert mock_client.chat.completions.create.called
    call_args = mock_client.chat.completions.create.call_args
    messages_sent = call_args[1]['messages']
    assert 'Bonjour' in messages_sent[0]['content']


@pytest.mark.parametrize('user_lang,expected_greeting_part', [
    ('fr', 'Bonjour'),
    ('es', 'Hola'),
    ('de', 'Hallo'),
    ('ja', 'こんにちは'),
    ('en', 'Hello'),
])
def test_language_selection_parametrized(i18n_loom, user_lang, expected_greeting_part):
    '''Test language selection for multiple languages'''
    greeting = i18n_loom['greeting_multilang']

    if user_lang != 'en':
        greeting_text = greeting.in_lang(user_lang) or str(greeting)
    else:
        greeting_text = str(greeting)

    assert expected_greeting_part in greeting_text


def test_i18n_context_usage(i18n_loom):
    '''Test using the i18n context in a prompt template'''
    context = i18n_loom['i18n_context']
    system_prompt = i18n_loom['davinci3_instruct_system']

    # These would be used together in a real scenario
    assert 'Internationalization' in str(context)
    assert 'i18n' in str(context)
    assert 'Obey the instruction' in str(system_prompt)


if __name__ == '__main__':
    raise SystemExit('Attention! Run with pytest')



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_openai_integration.py
'''
Test WordLoom integration with OpenAI API using mocks

pytest test/test_openai_integration.py
'''

import pytest

import wordloom


# Sample TOML content for OpenAI-style prompts
OPENAI_PROMPTS_TOML = b"""
# OpenAI-style prompts
lang = 'en'

[system_instruction]
_ = 'You are a helpful assistant that provides concise and accurate answers.'

[code_review_prompt]
_ = '''
Review the following code and provide feedback on:
1. Code quality and readability
2. Potential bugs or issues
3. Suggestions for improvement

Code:
{code_snippet}
'''
_m = ['code_snippet']

[translation_prompt]
_ = 'Translate the following text to {target_lang}: {text}'
_m = ['target_lang', 'text']

[summarize_prompt]
_ = 'Summarize the following text in {num_sentences} sentences:\\n\\n{text}'
_m = ['num_sentences', 'text']
"""


@pytest.fixture
def openai_loom():
    '''Load the OpenAI prompts loom'''
    return wordloom.load(OPENAI_PROMPTS_TOML)


def test_load_openai_prompts(openai_loom):
    '''Test that OpenAI-style prompts load correctly'''
    assert 'system_instruction' in openai_loom
    assert 'code_review_prompt' in openai_loom
    assert 'translation_prompt' in openai_loom

    # Check markers are preserved
    assert openai_loom['code_review_prompt'].markers == ['code_snippet']
    assert openai_loom['translation_prompt'].markers == ['target_lang', 'text']


def test_basic_template_expansion(openai_loom):
    '''Test basic template marker expansion'''
    prompt = openai_loom['translation_prompt']

    # Manual template expansion (simple example)
    expanded = str(prompt).format(target_lang='French', text='Hello, world!')
    assert 'French' in expanded
    assert 'Hello, world!' in expanded


def test_openai_chat_completion_format(openai_loom):
    '''Test formatting prompts for OpenAI chat completion API'''
    system_prompt = openai_loom['system_instruction']
    user_prompt = openai_loom['summarize_prompt']

    # Format for OpenAI API
    messages = [
        {'role': 'system', 'content': str(system_prompt)},
        {'role': 'user', 'content': str(user_prompt).format(
            num_sentences='3',
            text='This is a long article about AI. It discusses many topics...'
        )}
    ]

    assert messages[0]['role'] == 'system'
    assert 'helpful assistant' in messages[0]['content']
    assert messages[1]['role'] == 'user'
    assert '3 sentences' in messages[1]['content']


@pytest.mark.parametrize('target_lang,text,expected_in_output', [
    ('Spanish', 'Good morning', 'Spanish'),
    ('German', 'Thank you', 'German'),
    ('Japanese', 'Welcome', 'Japanese'),
])
def test_translation_prompt_variations(openai_loom, target_lang, text, expected_in_output):
    '''Test translation prompt with different languages'''
    prompt = openai_loom['translation_prompt']
    expanded = str(prompt).format(target_lang=target_lang, text=text)

    assert expected_in_output in expanded
    assert text in expanded


def test_mock_openai_completion(openai_loom, mocker):
    '''Test with a mocked OpenAI API call'''
    # Mock the OpenAI client
    mock_client = mocker.MagicMock()
    mock_response = mocker.MagicMock()
    mock_response.choices = [mocker.MagicMock()]
    mock_response.choices[0].message.content = 'This is a test response from the AI.'

    mock_client.chat.completions.create.return_value = mock_response

    # Use WordLoom prompt
    system_prompt = openai_loom['system_instruction']
    user_prompt = openai_loom['summarize_prompt']

    # Simulate API call
    response = mock_client.chat.completions.create(
        model='gpt-4',
        messages=[
            {'role': 'system', 'content': str(system_prompt)},
            {'role': 'user', 'content': str(user_prompt).format(
                num_sentences='2',
                text='Sample text to summarize.'
            )}
        ]
    )

    # Verify the mock was called correctly
    assert mock_client.chat.completions.create.called
    assert response.choices[0].message.content == 'This is a test response from the AI.'

    # Verify the prompt was formatted correctly
    call_args = mock_client.chat.completions.create.call_args
    messages_sent = call_args[1]['messages']
    assert len(messages_sent) == 2
    assert messages_sent[0]['role'] == 'system'
    assert messages_sent[1]['role'] == 'user'
    assert '2 sentences' in messages_sent[1]['content']


def test_code_review_prompt_expansion(openai_loom):
    '''Test code review prompt with actual code snippet'''
    prompt = openai_loom['code_review_prompt']

    code_snippet = '''
def add_numbers(a, b):
    return a + b
'''

    expanded = str(prompt).format(code_snippet=code_snippet)

    assert 'Code quality' in expanded
    assert 'def add_numbers' in expanded
    assert 'return a + b' in expanded


def test_metadata_preservation(openai_loom):
    '''Test that metadata is preserved from TOML'''
    # All prompts should have the default language
    assert openai_loom['system_instruction'].lang == 'en'
    assert openai_loom['code_review_prompt'].lang == 'en'


if __name__ == '__main__':
    raise SystemExit('Attention! Run with pytest')
